float Script : STANDARDSGLOBAL <
   string   UIWidget = "none";
   string   ScriptClass = "scene";
   string   ScriptOrder = "standard";
   string   ScriptOutput = "color";
   string   Script = "Technique=Fresnel_Spherical;";
   string   ScriptDescription = "<b>Fresnel Shader that uses a spherical environment map</b>";
> = 0.1; // version #

////////////////////////////////////////////////////////////
// TWEAKABLES

float fresnelPower
<
   string   UiLabel = "Fresnel Falloff";
   string   UiControl = "Slider";
  float UiMin = 0.0f;
  float UiMax = 10.0f;
> = 3.0;

float fresnelStrength
<
   string   UiLabel = "Frensel Strength";
   string   UiControl = "Slider";
  float UiMin = 0.0f;
  float UiMax = 10.0f;
> = 1.0;

float normalStrength
<
   string   UiLabel = "Normal Strength";
   string   UiControl = "Slider";
  float UiMin = 0.0f;
  float UiMax = 10.0f;
> = 1.0;

float environmentLighting
<
   string   UiLabel = "Ambient Environment";
   string   UiControl = "Slider";
  float UiMin = 0.0f;
  float UiMax = 10.0f;
> = 0.0;

float envRotation
<
   string   UiName = "Env Rotation";
   string   UiControl = "Slider";
  float UiMin = -99999999999.0f;
  float UiMax = 99999999999.0f;
> = 0;

float4 cameraPositionWS <   string BindAddress = "Tog.Camera.Location";>;
float4x4 worldMatrix <    string BindAddress = "Tog.Camera.Model"; >;
float4x4 invTransWorldMatrix <   string BindAddress = "Tog.Camera.InverseTransposeModel";>;
float4x4 inverseViewMatrix <   string BindAddress = "Tog.Camera.InverseView";>;

////////////////////////////////////////////////////////////
// TYPEDEFS

struct TogAmbientLight
{
  float4 ambient;
};

struct TogDirectionalLight
{
  float3 ambient;
  float3 diffuse;
  float3 specular;
  float4 position;
};

struct TogPointLight
{
  float3 ambient;
  float3 diffuse;
  float3 specular;
  float4 position;
  float3 attenuation;
};

struct TogSpotLight
{
  float3 ambient;
  float3 diffuse;
  float3 specular;
  float3 direction;
  float4 position;
  float3 expCutOff;
  float3 attenuation;
};

struct TogMaterial
{
  float3 ambient;
  float4 diffuse;
  float3 specular;
  float3 emission;
  int    shininess;
};

struct TogTexGen
{
  float4 plane;
  float  enabled;
  float  mode;
};

////////////////////////////////////////////////////////////////

float attenuate(float3 a,float3 b,float3 atten)
{
  float d = distance(a,b);
  return 1.0f / (atten.x + atten.y*d + atten.z*d*d);
}

////////////////////////////////////////////////////////////////

float spotEffect(float3 L,float3 D,float cutOff,int exponent)
{
  float ang = max(dot(normalize(L),normalize(D)),0.0f);
  if(ang > cutOff)
  return pow(ang,exponent);
  else
  return 0.0f;
}

////////////////////////////////////////////////////////////////

float dualConeSpotEffect(float3 L,float3 D,float cutOff,float outerCutOff,int exponent)
{
  float ang = max(dot(normalize(L),normalize(D)),0.0f);

  return pow(smoothstep(outerCutOff,cutOff,ang), exponent);
}

////////////////////////////////////////////////////////////////

float2 modf(float2 a)
{
  return a-floor(a);
}

////////////////////////////////////////////////////////////////

float linstep(float min,float max,float v)
{
  return clamp((v-min)/(max-min),0,1);
}

////////////////////////////////////////////////////////////////

float ReduceLightBleeding(float p_max,float amount)
{
  return linstep(amount,1,p_max);
}

////////////////////////////////////////////////////////////////

float ChebyshevUpperBound(float2 moments,float t)
{
  float p = (t<=moments.x);

  float var = moments.y - (moments.x*moments.x);
  var = max(var,0.0) + 0.01;
  var = min(var,1.0);

  float d = moments.x-t;
  float p_max = var / (var + d*d);
  float l = max(p,p_max);

  return smoothstep(0.4,1.0,l);
}

////////////////////////////////////////////////////////////////
// EXTERNAL DATA

uniform int lightingEnabled         < string BindAddress = "Tog.Lighting.Enabled"; >;
uniform float4  clip04Enabled       < string BindAddress = "Tog.ClipPlane04Enabled"; >;
uniform float4  clip04SoftDistance  < string BindAddress = "Tog.ClipPlane04SoftDistance"; >;

float4x4 WorldInverseTranspose      < string BindAddress = "Tog.Camera.InverseTransposeModel"; >;
float4x4 WorldViewProjection        < string BindAddress = "Tog.Camera.ModelViewProjection";>;
float4x4 World                      < string BindAddress = "Tog.Camera.Model";>;
float4x4 ViewInverse                < string BindAddress = "Tog.Camera.InverseView";>;

////////////////////////////////////////////////////////////////
//texture samplers

// Decal Texture
sampler2D tex0                      < string BindAddress = "Tog.Texture[0].2D"; >;
float tex0ColourSpace               < string BindAddress = "Tog.Texture[0].ColourSpace"; >;
float4x4 tex0Matrix                 < string BindAddress = "Tog.Texture[0].Matrix"; >;

// SpecGlossAO (rgb)
sampler2D tex1                      < string BindAddress="Tog.Texture[1].2D"; >;
float4x4 tex1Matrix                 < string BindAddress = "Tog.Texture[1].Matrix"; >;

// Environment
sampler2D tex2                      < string BindAddress="Tog.Texture[2].2D"; >;

// Normal
sampler2D tex3                      < string BindAddress="Tog.Texture[3].2D"; >;

// We also need some overloaded values for the Fresnel ... so we'll use emission
float3 overloadedElements           < string BindAddress = "Tog.Material.Emission"; >;


////////////////////////////////////////////////////////////////
//application to vertex structure

struct VertIn
{
  float4 position   : POSITION;//4 -> 3
  float3 normal     : NORMAL;
  float3 tangent : TANGENT;
  float3 bitangent : BINORMAL;
  float4 colour     : COLOR;
  float4 texCoord0  : TEXCOORD0;
  float4 texCoord1  : TEXCOORD1;
};

////////////////////////////////////////////////////////////////
//vertex to fragment structure

struct VertOut
{
  float4 position   : POSITION;
  float4 colour     : COLOR;
  float4 texCoord0  : TEXCOORD0;
  float4 texCoord1  : TEXCOORD1;
  float3 normal     : TEXCOORD2;
  float3 tangent : TEXCOORD3;
  float3 bitangent : TEXCOORD4; 
  float3 eyePos     : TEXCOORD5;
  float4 clipValues : TEXCOORD6;
};

////////////////////////////////////////////////////////////////
//the vertex program

VertOut vertex(
VertIn IN)
{
  VertOut OUT;      

  OUT.position = mul(glstate.matrix.mvp, IN.position);

  OUT.eyePos = IN.position;

  OUT.normal = normalize(IN.normal);
  OUT.tangent = normalize(IN.tangent);
  OUT.bitangent = normalize(IN.bitangent);

  OUT.colour = IN.colour;
  OUT.clipValues = 0.0;

  float4 eyePosVS = mul(glstate.matrix.modelview[0], IN.position);

  if(clip04Enabled.x == 1.0f)
    OUT.clipValues.x = dot(eyePosVS.xyz,glstate.clip[0].plane.xyz) + glstate.clip[0].plane.w;
  if(clip04Enabled.y == 1.0f)
    OUT.clipValues.y = dot(eyePosVS.xyz,glstate.clip[1].plane.xyz) + glstate.clip[1].plane.w;
  if(clip04Enabled.z == 1.0f)
    OUT.clipValues.z = dot(eyePosVS.xyz,glstate.clip[2].plane.xyz) + glstate.clip[2].plane.w;
  if(clip04Enabled.w == 1.0f)
    OUT.clipValues.w = dot(eyePosVS.xyz,glstate.clip[3].plane.xyz) + glstate.clip[3].plane.w;

  float4 texCoord0 = IN.texCoord0; //2D UVs

  OUT.texCoord0 = mul(tex0Matrix,texCoord0);

  float4 texCoord1 = IN.texCoord1; //Cube UVs
  {
    float3 Vr = normalize(OUT.eyePos);
    float3 Rr = -reflect(Vr,OUT.normal);
    texCoord1.x = Rr.x;
    texCoord1.y = Rr.y;
    texCoord1.z = Rr.z;
  }

  OUT.texCoord1 = mul(tex1Matrix,texCoord1);

  return OUT;
}

////////////////////////////////////////////////////////////////
//spherical mapping

float3 envSphereSample(
float4 dirWS)
{
  float n = length(dirWS.xz);
  float2 pos = float2( (n>0.0000001) ? dirWS.x / n : 0.0, dirWS.y);
  pos = acos(pos)*0.31830988618379067153776752674503;
  pos.x = (dirWS.z > 0.0) ? pos.x*0.5 : 1.0-(pos.x*0.5);
  pos.x += envRotation/360;
  pos.y = 1.0-pos.y;
  // return tex2D(tex2, pos).rgb;

  return tex2Dlod(tex2, float4(pos.xy,1,dirWS.w)).rgb;
}

////////////////////////////////////////////////////////////////
//the pixel program

float4 pixel(
VertOut IN,
uniform sampler2D tex0,
uniform sampler2D tex1,
uniform sampler2D tex2,
uniform sampler2D tex3
) : COLOR
{
  ////////////////////////////////////////////////////////////////
  //the combined texture colours
  float4 decalColour = {1.0f,1.0f,1.0f,1.0f};

  //the overall diffuse and ambient lighting terms
  float3 lightColour = {0.0f,0.0f,0.0f};

  //the overall specular lighting term
  float3 lightSpecular = {0.0f,0.0f,0.0f};

  ////////////////////////////////////////////////////////////////
  //normal and eye vector

  float3 worldNormal= normalize(mul((float3x3)glstate.matrix.invtrans.modelview[0], IN.normal));

  //gets normal map value for this texel, transforms it to tangent space
  float3 normalMap = tex2D(tex3, IN.texCoord0.xy).xyz;
  normalMap = (normalMap.xyz - 0.5) * 2.0;

  // "The following line includes NormalStrength parameter. This was borrowed from Mark's ReflectioNBump code. 
  // Not sure whether it's appropriate." -Seb April 2015
  normalMap = normalize(normalMap*float3(normalStrength, normalStrength,1));

  // Back to Jesse's good code ..
  float3 normalMapTS = normalize( normalMap.x * IN.tangent
                                  + normalMap.y * IN.bitangent
                                  + normalMap.z * IN.normal);

  float3 normalMapWS = normalize(mul((float3x3)invTransWorldMatrix, normalMapTS));

  float3 E = normalize(mul(glstate.matrix.modelview[0], float4(IN.eyePos, 1)));
  float3 EWS = normalize(float3(cameraPositionWS - (mul(worldMatrix, float4(IN.eyePos,1)))).xyz);

  float4 specMap = float4(tex2D(tex1,IN.texCoord0.xy).rrr,1);

  float shininess = glstate.material.front.shininess;

  ////////////////////////////////////////////////////////////////
  //Clipping + Soft Clipping
  float clipAlpha = 1.0f;
  if(any(clip04Enabled))
  {
    if(any(clip04SoftDistance))
    {
      IN.clipValues += max(-clip04SoftDistance,0);
      float4 clips;
      clips = IN.clipValues*(1.0/abs(clip04SoftDistance));
      clipAlpha = clamp(min(min(min(clips.x,clips.y),clips.z),clips.w),0,1);
    }
    clip(IN.clipValues);
  }

  float4 decalModulate;

  ////////////////////////////////////////////////////////////////
  //Add maps

  float4 decalTemp;

  // Handle RGB/YUV colour spaces
  if (tex0ColourSpace == 0) // RGB
  {
    decalTemp = tex2D(tex0,IN.texCoord0.xy);
  }
  else 
  {
    float2  texLum,texCB,texCR,texCoord;
    float4  tex;

    texCoord = modf(IN.texCoord0.xy);

    if(tex0ColourSpace == 1) // YUV420
    {    
      texLum = float2(texCoord.x,texCoord.y*2.0/3.0);
      texCB = float2(texCoord.x*0.5,texCoord.y/3.0 + (2.0/3.0));
      texCR = float2(texCB.x+0.5,texCB.y);
    }
    else // YUV422
    {
      texLum = float2(texCoord.x,texCoord.y/2.0);
      texCB = float2(texCoord.x*0.5,texCoord.y/2.0 + 0.5);
      texCR = float2(texCB.x+0.5,texCB.y);
    }

    tex.r = tex2D(tex0, texLum);
    tex.b = tex2D(tex0, texCB);
    tex.g = tex2D(tex0, texCR);
    tex.a = 1.0;

    tex = tex + float4(0, -0.5, -0.5, 0);

    decalTemp = float4( tex.r + (tex.g * 1.4022),
                        tex.r + (tex.b * -0.3456) + (tex.g * -0.7145),
                        tex.r + (tex.b * 1.7710),
                        1.0);
  }

  decalModulate = (float4(glstate.material.diffuse.rgb,1)*decalTemp)+(float4(lightColour,1)*decalTemp)+float4(glstate.material.ambient.rgb,1);

  {
    float3 reflectEnviroMap;
    float3 eyeWS = normalize(mul(inverseViewMatrix, float4(0,0,0,1)) - mul(worldMatrix, float4(IN.eyePos.xyz, 1))).xyz;
    reflectEnviroMap = -normalize(reflect(eyeWS,normalMapWS));

    float4 uvw = float4(reflectEnviroMap,1);
    float4 uvw_blur = float4(reflectEnviroMap,11);

    float4 enviroMap;

    {
      enviroMap = float4(envSphereSample(uvw), 4);
    }

    float4 enviroAmbient;

    {
      enviroAmbient = float4(envSphereSample(uvw_blur), 4);
    }

    // fresnelLighting

    float3 fresnelLighting;
    fresnelLighting = (pow((1-abs(dot(EWS, normalMapWS))),fresnelPower))*fresnelStrength;


    // spec highlights

    float4 specLighting;
    specLighting = float4(lightSpecular,1)*specMap;

    // enviroLighting

    float4 environmentalSpecLighting;
    environmentalSpecLighting = enviroMap*specMap;

    float4 totalFresnel = float4(fresnelLighting,1)*enviroMap*specMap;
    decalModulate += totalFresnel + specLighting + (environmentLighting*enviroAmbient); 
  }

  decalModulate += float4(glstate.material.emission.rgb,0);

  return float4(decalModulate.rgb, decalTemp.a*glstate.material.front.diffuse.a*clipAlpha);
}

////////////////////////////////////////////////////////////////
//vertex to fragment structure
technique Fresnel_Spherical
{
  pass p0
  {
    VertexProgram = compile vp40 vertex();
    FragmentProgram = compile fp40 pixel(tex0,tex1,tex2,tex3);
  }
}

